import { useMemo, useState } from 'react';
import { ConnectingNodeDataI } from '../utils/connectorDrop';
import { useActiveQuery, useActiveSaveState, useSchemaGraph } from '@/lib/data-access';
import { useDispatch } from 'react-redux';
import { toSchemaGraphology } from '@/lib/data-access/store/schemaSlice';
import { SchemaNode, QueryElementTypes, Handles, SchemaEdge, defaultGraph } from 'ts-common';
import { toHandleData } from '../../model';
import { setQuerybuilderGraphology, toQuerybuilderGraphology } from '@/lib/data-access/store/sessionSlice';
import { Button, EntityPill, RelationPill } from '@/lib/components';
import { DatabaseType } from 'ts-common/src/model/webSocket/dbConnection';
import { edge2schemaEdge, nodeEdges, relationEdgeIds } from './schema2queryUtils';

export const QueryBuilderRelatedNodesPanel = (props: {
  reactFlowWrapper: HTMLDivElement | null;
  className?: string;
  title?: string;
  onFinished: () => void;
  connection?: ConnectingNodeDataI | null;
}) => {
  const schema = useSchemaGraph();
  const activeQuery = useActiveQuery();
  const activeSS = useActiveSaveState();
  const dispatch = useDispatch();
  const graphologyGraph = toQuerybuilderGraphology(activeQuery?.graph || defaultGraph());
  const schemaGraph = toSchemaGraphology(schema);
  const [selectedType, setSelectedType] = useState<'entity' | 'relation' | 'all'>('all');

  const dbType: DatabaseType = activeSS?.dbConnections?.[0]?.type ?? DatabaseType.Neo4j;
  const handleData = props.connection?.params?.handleId ? toHandleData(props.connection.params.handleId) : null;

  const relatedEntities = useMemo<SchemaNode[]>(() => {
    if (!props.connection) return [];
    const type = props.connection.node.type;
    if (type !== QueryElementTypes.Entity && type !== QueryElementTypes.Relation)
      throw new Error('Connection is not an entity or relation');

    if (type === QueryElementTypes.Entity) {
      // find entities that are connected to the current entity
      const nodes = schemaGraph.neighbors(props.connection.node.schemaKey);
      return nodes.map(node => schemaGraph.getNodeAttributes(node) as SchemaNode);
    } else if (type === QueryElementTypes.Relation) {
      // find relations of the current entity
      const attributes = schemaGraph.getEdgeAttributes(props.connection.node.schemaKey);
      return [schemaGraph.getNodeAttributes(attributes.to), schemaGraph.getNodeAttributes(attributes.from)];
    }

    return [];
  }, [schema, activeQuery?.graph, props]);

  const relatedRelations = useMemo<SchemaEdge[]>(() => {
    if (!props.connection || !props.connection.node.schemaKey) return [];
    return nodeEdges(props.connection.node, schemaGraph);
  }, [schema, activeQuery?.graph, props]);

  const newEntity = (entity: SchemaNode) => {
    if (!props.connection) {
      props.onFinished();
      return;
    }

    const params = props.connection.params;
    const position = props.connection.position;

    const newNode = graphologyGraph.addPill2Graphology(
      {
        type: QueryElementTypes.Entity,
        x: position.x,
        y: position.y,
        name: entity.name,
        schemaKey: entity.name,
        attributes: [],
      },
      schemaGraph.getNodeAttribute(entity.name, 'attributes'),
    );

    if (!newNode?.id) throw new Error('Logic node has no id');
    if (!newNode?.name) throw new Error('Logic node has no name');
    if (!params.handleId) throw new Error('Connection has no source or target');

    const originalNodeId = graphologyGraph.getNodeAttributes(params.nodeId);

    if (dbType === DatabaseType.Postgres) {
      // Postgres does not support entity to entity connection, so we need to add a relation in between
      if (!props?.connection?.node?.name) return;

      const schemaRelation = nodeEdges(props.connection.node, schemaGraph).find(
        edge => edge.to === entity.name || edge.from === entity.name,
      );
      if (!schemaRelation) return;

      const relationNode = graphologyGraph.addPill2Graphology(
        {
          type: QueryElementTypes.Relation,
          x: position.x,
          y: position.y,
          schemaKey: schemaRelation.label,
          depth: { min: activeQuery?.settings.depth.min || 1, max: activeQuery?.settings.depth.max || -1 },
          name: schemaRelation.collection,
          collection: schemaRelation.collection,
          attributes: [],
        },
        schemaGraph.getEdgeAttribute(schemaRelation.label, 'attributes'),
      );

      if (handleData?.handleType === Handles.EntityRight || handleData?.handleType === Handles.RelationRight)
        graphologyGraph.addEdge2Graphology(originalNodeId, relationNode);
      else graphologyGraph.addEdge2Graphology(relationNode, originalNodeId);
      if (handleData?.handleType === Handles.EntityRight || handleData?.handleType === Handles.RelationRight)
        graphologyGraph.addEdge2Graphology(relationNode, newNode);
      else graphologyGraph.addEdge2Graphology(newNode, relationNode);
    } else {
      if (handleData?.handleType === Handles.EntityRight || handleData?.handleType === Handles.RelationRight)
        graphologyGraph.addEdge2Graphology(originalNodeId, newNode);
      else graphologyGraph.addEdge2Graphology(newNode, originalNodeId);
    }

    dispatch(setQuerybuilderGraphology(graphologyGraph));
    props.onFinished();
  };
  const newRelation = (relation: SchemaEdge) => {
    if (!props.connection) {
      props.onFinished();
      return;
    }

    const params = props.connection.params;
    const position = props.connection.position;

    const newNode = graphologyGraph.addPill2Graphology(
      {
        type: QueryElementTypes.Relation,
        x: position.x,
        y: position.y,
        schemaKey: relation.label,
        depth: { min: activeQuery?.settings.depth.min || 1, max: activeQuery?.settings.depth.max || -1 },
        name: relation.collection,
        collection: relation.collection,
        attributes: [],
      },
      schemaGraph.getEdgeAttribute(relation.label, 'attributes'),
    );

    if (!newNode?.id) throw new Error('Logic node has no id');
    if (!newNode?.name) throw new Error('Logic node has no name');
    if (!params.handleId) throw new Error('Connection has no source or target');

    if (handleData?.handleType === Handles.EntityRight || handleData?.handleType === Handles.RelationRight)
      graphologyGraph.addEdge2Graphology(graphologyGraph.getNodeAttributes(params.nodeId), newNode);
    else graphologyGraph.addEdge2Graphology(newNode, graphologyGraph.getNodeAttributes(params.nodeId));
    dispatch(setQuerybuilderGraphology(graphologyGraph));
    props.onFinished();
  };

  return (
    <div className={props.className + ' card'}>
      {props.title && <h1 className="card-title mb-7">{props.title}</h1>}
      <div className="overflow-x-hidden h-[75rem] w-full gap-2 pt-3">
        <div className="w-full grid grid-cols-2 gap-2">
          <Button
            label="Related Entities"
            variant={selectedType === 'entity' ? 'solid' : 'outline'}
            size="2xl"
            className="w-full"
            onClick={e => {
              e.preventDefault();
              if (selectedType === 'entity') setSelectedType('all');
              else setSelectedType('entity');
            }}
          />
          <Button
            label="Related Relations"
            variant={selectedType === 'relation' ? 'solid' : 'outline'}
            size="2xl"
            className="w-full"
            onClick={e => {
              e.preventDefault();
              if (selectedType === 'relation') setSelectedType('all');
              else setSelectedType('relation');
            }}
          />
          {selectedType !== 'relation' ? (
            <ul className="menu p-0 [&_li>*]:rounded-none pb-10">
              {relatedEntities.map(entity => {
                return (
                  <Button key={entity.name} onClick={() => newEntity(entity)} className="w-fit p-2" variant="ghost" tooltip={entity.name}>
                    <EntityPill title={entity.name} pillWidth={250} align="center" />
                  </Button>
                );
              })}
            </ul>
          ) : (
            <span></span>
          )}
          {selectedType !== 'entity' ? (
            <ul className="menu p-0 [&_li>*]:rounded-none w-full pb-10">
              {relatedRelations.map(relation => {
                return (
                  <Button
                    key={relation.name}
                    onClick={() => newRelation(relation)}
                    className="w-fit p-2"
                    variant="ghost"
                    tooltip={relation.name}
                  >
                    <RelationPill title={relation.name} pillWidth={250} align="center" />
                  </Button>
                );
              })}
            </ul>
          ) : (
            <span></span>
          )}
        </div>
      </div>
    </div>
  );
};
